﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using System.Windows;
using Windows.Devices.Bluetooth.Advertisement;
using Windows.Storage.Streams;

namespace iBeaconScanner
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        private BluetoothLEAdvertisementWatcher advWatcher;

        /*　定数　*/
        public static int CompanyID_Apple = 76;
        public static int Length_of_iBeacon_Advertise_Packet = 23;

        /* パラメータ */
        public static int A_thresh = 2;//認証可能人数
        public static double B_thresh = 0.3;//RU率
        public static TimeSpan LIMIT_TIME = new TimeSpan(0, 0, 30);//n秒パケットが取れなければ終了
        public static short RSSI = -70;//探索範囲 大きいほど狭い

        static int RUcnt, NUcnt;
        double rate;
        
        Stopwatch sw = new Stopwatch();

        static HashSet<string> RU_uuid = new HashSet<string>(); //RUのハッシュセット
        static private HashSet<string> All_uuid = new HashSet<string>(); //RUのハッシュセット

     
        public MainWindow()
        {
            InitializeComponent();                   

            /* 事前登録 */
            RU_uuid.Add("48534442-4c45-4144-1234-dcba12345678");
            RU_uuid.Add("48534442-4c45-4144-1234-abcd12345678");
            RU_uuid.Add("48534442-4c45-4144-1234-abcd12345671");//me

            RU_uuid.Add("48534442-4c45-4144-80c0-1800ffffffff");//deafault
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff0");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff1");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff2");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff3");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff4");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff5");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff6");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff7");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff8");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffff9");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffffa");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffffb");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffffc");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffffd");
            RU_uuid.Add("48534442-4c45-4144-80c0-1800fffffffe");
            

            this.para.Text = "A="+ A_thresh + ",B=" + B_thresh + ",L="+LIMIT_TIME + ",RSSI="+RSSI;
        
            this.advWatcher = new BluetoothLEAdvertisementWatcher();

            // https://blogs.msdn.microsoft.com/shozoa/2016/02/28/windows-10-bluetooth/
            // インターバルがゼロのままだと、CPU負荷が高い。適切な間隔(SDK サンプルでは 1秒)に指定しないと、アプリの動作に支障をきたすことになる。
            this.advWatcher.SignalStrengthFilter.SamplingInterval = TimeSpan.FromMilliseconds(1000);

            // rssi >= RSSIのときスキャンする
            this.advWatcher.SignalStrengthFilter.InRangeThresholdInDBm = RSSI;//-50 で3m強 -70なら研究室の端から端までOK

            this.advWatcher.Received += this.Watcher_Received;

            sw.Start();

            // スキャン開始
            this.advWatcher.Start();

        }

        private static Mutex mutex = new Mutex();

        private async void Watcher_Received(BluetoothLEAdvertisementWatcher sender, BluetoothLEAdvertisementReceivedEventArgs args)
        {
                  
            await this.Dispatcher.InvokeAsync(() =>
            {

                //mutex.WaitOne();

                Debug.WriteLine($"{sw.Elapsed}");//経過時間
                if (sw.Elapsed > LIMIT_TIME)
                {
                    Debug.WriteLine($"#############");
                    advWatcher.Stop();
                }

                foreach (var manufacturerData in args.Advertisement.ManufacturerData)
                {
                     mutex.WaitOne();

                    Debug.WriteLine($"-------------");
                    Debug.WriteLine($"???{args.Timestamp} RSSI={args.RawSignalStrengthInDBm} BD:{args.BluetoothAddress}");


                    if (manufacturerData.CompanyId != CompanyID_Apple)
                    {
                        //Debug.WriteLine($"---NOT Apple's packet");
                        continue;
                    }
                    if (manufacturerData.Data.Length != Length_of_iBeacon_Advertise_Packet)
                    {
                        //Debug.WriteLine($"---NOT iBeacon packet");
                        continue;
                    }

                   
                   // Debug.WriteLine($"{args.Timestamp} RSSI={args.RawSignalStrengthInDBm} BD:{args.BluetoothAddress}");
                    // Debug.WriteLine($"{args.AdvertisementType}, {manufacturerData.CompanyId}");

                    /* All_uuidリストを作成するフェーズ ↓*/
                    using (var reader = DataReader.FromBuffer(manufacturerData.Data))
                    {
                        reader.ByteOrder = ByteOrder.BigEndian;// BigEndian:順方向にパケットを保存


                        /* 0215 で始まればiBeacon */  //http://d.hatena.ne.jp/shu223/20141208/1418051009
                        var indicator = reader.ReadByte();
                        var length = reader.ReadByte();
                        if (indicator != 0x02 && length != 0x15)
                        {
                            Debug.WriteLine($"  ?1");
                            continue; //iBeacon でないAdvertise Packet
                        }

                       // Debug.WriteLine("   This is iBeacon Packet!!!");

                        // Parse advertise packet.
                        var uuidA = reader.ReadInt32();
                        var uuidB = reader.ReadInt16();
                        var uuidC = reader.ReadInt16();
                        var uuidD = new byte[8];
                        reader.ReadBytes(uuidD);
                        var major = reader.ReadUInt16();
                        var minor = reader.ReadUInt16();
                        var power = reader.ReadByte();

                        var proximity_uuid = new Guid(uuidA, uuidB, uuidC, uuidD);

                        Debug.WriteLine($"  UUID:{proximity_uuid}, major: {major}, minor: {minor}, power: {power}");

                        /* 周囲のuuidの hash set を作成 */
                        if (!All_uuid.Contains(proximity_uuid.ToString())) //新しいUUIDなら追加
                        {
                            All_uuid.Add(proximity_uuid.ToString());
                            Debug.WriteLine($"  It's New UUID!");
                            this.listBox.Text = this.listBox.Text +args.Timestamp +"///"+ proximity_uuid + "\r\n";
                        }
                    }
                    /* All_uuidリストを作成するフェーズ ↑*/

                    //mutex.ReleaseMutex();
                }

                /* 判定フェーズ↓ */
                RUcnt = 0;
                foreach (string N in RU_uuid) //RUリストが大きいときは逆の入れ子の方がいいかも
                {
                    if (All_uuid.Contains(N))
                    {
                        RUcnt++;
                        continue;
                    }
                }
                NUcnt = All_uuid.Count - RUcnt;

                this.RUNUBox.Text = "RU:" + RUcnt + "人,　NU:" + NUcnt + "人" + "\r\n";

                rate = (1.0 * RUcnt / (RUcnt + NUcnt));
                RUrate.Text = " RUrate=" + rate.ToString("F2");

                if (RUcnt >= A_thresh &&  rate >= B_thresh)//安全かどうかの判定
                {
                    this.textBox.Text = "You are ACCEPTED!!";
                    Debug.WriteLine($"  You are ACCEPTED!!");
                }
                else
                {
                    this.textBox.Text = "Input Password";
                    Debug.WriteLine($"  Input Password");
                }
                /* 判定フェーズ↑ */

            });   
        }
    }
}
